require("sound.lua")
require("enemytypes.lua")

function love.run()
  _r = math.random
  _f = math.floor
  _c = math.cos
  _s = math.sin
  _sq = math.sqrt
  _at2 = math.atan2
  _d2r = math.rad
  _r2d = math.deg
  _m = math.mod
  pi = math.pi
  gr = love.graphics
  ti = love.timer
  kb = love.keyboard
  math.randomseed(os.time()) _r() _r() _r() _r()

	love.load(arg)
	local dt = 0
	while true do
	  dt = ti.getDelta()
	  time = ti.getTime()
		ti.step()
		gr.clear()
		love.update(dt)
		love.draw(dt)
		if love.event then
			for e,a,b,c in love.event.poll() do
				if e == "q" then
					if love.audio then love.audio.stop() end
				return
				end
				love.handlers[e](a,b,c)
			end
		end
		ti.sleep(1)
		gr.present()
	end
end

function love.load()
  bg = {r=0, g=0, b=0, rt=0, gt=0, bt=0, speed = 256*2}

  soundOn = false
  debugVar = false

  screenW = gr.getWidth()
  screenH = gr.getHeight()

	--muzak = {}
	--muzak["bgm"] = love.audio.newSource("audio/bgm.ogg", "stream")
  --bgm = love.audio.newSource("audio/bgm.ogg", "stream")

	sfx = {}
	sfx["player_fire_1"] = love.sound.newSoundData("audio/97822__phisto99__FS_021_Laser_or_shoot.wav")
	sfx["player_fire_5"] = love.sound.newSoundData("audio/97820__phisto99__FS_019_Laser_or_shoot.wav")
	sfx["enemy_fire_1"] = love.sound.newSoundData("audio/97821__phisto99__FS_020_Laser_or_shoot.wav")
	sfx["explode_1"] = love.sound.newSoundData("audio/31874__HardPCM__Chip037.wav")
	sfx["explode_2"] = love.sound.newSoundData("audio/33215__HardPCM__Chip075.wav")
	sfx["splash_1"] = love.sound.newSoundData("audio/9508__petenice__SPLASH.wav")
	sfx["powerup_1"] = love.sound.newSoundData("audio/97824__phisto99__FS_023_Pickup_or_coin.wav")

  achievements = {}
  achievements["test"] = { desc = "This is a sample achievement.", unlocked = false, secret = false, value = 10 }
  achievements["mrgreen"] = { desc = "ACHIEVEMENT UNLOCKED: MEAN MR. GREEN", unlocked = false, secret = false, value = 10 }

  screenScale = (screenW/800)

  lw = math.max(1*screenScale,1.5)
  loadAlphabet()
  pauseColor = {255,255,255}

  timeout = {}

  setupMenu()
end

function love.update(dt)
  updateBG(dt)
  updateTimeout(dt)

  local kLeft = kb.isDown("left") or kb.isDown("a")
  local kRight = kb.isDown("right") or kb.isDown("d")
  local kUp = kb.isDown("up") or kb.isDown("w")
  local kDown = kb.isDown("down") or kb.isDown("s")
  local kSpace = kb.isDown(" ")
  local kShift = kb.isDown("lshift")

  if inMenu then

  else
    if level.readyMessage > 0 then
      level.readyMessage = level.readyMessage - dt
    end

    if paused then

    else
      --level.speed = level.speed + 10*dt
      player.tileX = _f(player.x/32)
      level.scrollPause = level.scrollPause - dt
      if level.scrollPause <= 0 then
        level.offset = level.offset + level.speed * dt
        if level.offset > (level.length * 32) - _f(800/32) then
          level.offset = (level.length * 32) - _f(800/32)
          level.speed = 0
          level.speedTo = 0
        end
        if player.dying == false then player.x = player.x + level.speed * dt end
      end
      level.paraOffset = level.offset * .5

      player.inv = player.inv - dt
      if player.inv <= 0 then player.inv = 0 end

      level.sunmoon.y = level.sunmoon.y - 4 * dt

      if level.speed < level.speedTo then
        level.speed = level.speed + (64 * dt)
        if level.speed > level.speedTo then level.speed = level.speedTo end
      end

      if level.speed > level.speedTo then
        level.speed = level.speed - (64 * dt)
        if level.speed < level.speedTo then level.speed = level.speedTo end
      end

      if player.y-10 < level.map[player.tileX].ceil then
        if player.drifting == 0 then damagePlayer(10) end
        player.drifting = 1
        player.y = player.y + 20
      end
      if player.dying then
        player.hasShield = false
        if player.y > 500-level.map[player.tileX].floor then
          if player.dead == false then
            playSound("explode_1")
            for i=0,4 do
              setTimeout(function() spawnExplosion(player.x+_r(-1,1)*10,player.y,1) end, i*.2)
            end
            killPlayer()
          end
          player.dead = true
        end
        if player.y > 500-level.liquidHeight then
          if player.dead == false then
            playSound("splash_1")
            playSound("explode_1")
            for i=0,4 do
              setTimeout(function() spawnExplosion(player.x+_r(-1,1)*10,player.y,2) end, i*.2)
            end
            killPlayer()
          end
          player.dead = true
        end
      else
        if player.y+10 > 500-level.map[player.tileX].floor then
          if player.drifting == 0 then damagePlayer(10) end
          player.drifting = -1
          player.y = player.y - 20
        end
      end

      if player.hitFlash > 0 then player.hitFlash = player.hitFlash - dt end

      if player.hasShield then
        player.shieldTime = player.shieldTime - dt
        if player.shieldTime <= 0 then player.hasShield = false end
        if player.shieldRad < 48 then player.shieldRad = player.shieldRad + 256*dt if player.shieldRad > 48 then player.shieldRad = 48 end end
      else
        if player.shieldRad > 0 then player.shieldRad = player.shieldRad - 256*dt if player.shieldRad < 0 then player.shieldRad = 0 end end
      end

      player.timeShot = player.timeShot - dt
      if player.timeShot < 0 then player.timeShot = 0 end

      if player.dying == false then
        if kSpace or kShift then
          if player.timeShot <= 0 then
            playerFireWeapon()
            player.timeShot = player.timeBetween
          end
        end

        if kLeft then
          player.x = player.x - player.xs * dt
          if player.x < level.offset + 20 then player.x = level.offset + 20 end
        elseif kRight then
          player.x = player.x + player.xs * dt
          if player.x > level.offset + 780 then player.x = level.offset + 780 end
        end

        if kUp then
          if player.y > level.map[player.tileX].ceil + 10 then
            player.y = player.y - player.ys * dt
          end
        elseif kDown then
          if player.y < (500-level.map[player.tileX].floor) - 10 then
            player.y = player.y + player.ys * dt
          end
        end
      end

      for i, e in pairs(enemy) do
        if e.dead == false then
          local s = 96

          if distanceFrom(player.x, player.y, e.x, e.y) < (e.radius + 16) then
            damagePlayer(10)
          end

          if e.type >= 100 then
            e.doUpdate(i, dt)
          else
            e.tsb = e.tsb - dt
            if e.tsb < 0 and e.x < level.offset+750 then
              spawnBullet(e.x, e.y, 5, 180, 512, "enemy")
              playSound("enemy_fire_1")
              e.tsb = 1
            end

            if e.type == 2 then s = 128 end
            e.x = e.x - s * dt
            if e.type == 2 then
              e.y = e.y + _s(ti.getTime() * 2)
            end
          end

          if e.x < level.offset - 32 then
            e.dead = true
          end
          for j, b in pairs(bullet) do
            if b.dead == false then
              if b.x > level.offset + 820 or b.x < level.offset - 20 then
                b.dead = true
              end
              local bxt = math.min(_f(b.x/32),level.length-1)
              if b.y > (500-level.map[bxt].floor) then
                b.dead = true
                spawnExplosion(b.x,b.y,1)
              end
              if b.y > (500-level.liquidHeight) then
                b.dead = true
                spawnExplosion(b.x,b.y,2)
              end
            end
            if b.who == "player" then
              if distanceFrom(e.x, e.y, b.x, b.y) < e.radius and b.dead == false and b.dead == false then
                b.dead = true
                e.energy = e.energy - b.power
                if e.energy <= 0 then
                  if e.type >= 100 then
                    e.death(i)
                  end
                  if e.powerup > 0 then
                    powerup[#powerup+1] = { x = e.x, y = e.y, item = e.powerup, give = powerupList[e.powerup].give, dead = false }
                  end
                  playSound("explode_1")
                  e.dead = true
                  spawnExplosion(e.x,e.y,1)
                  player.score = player.score + e.points * player.multiplier
                  player.multiplierCounter = player.multiplierCounter + 1
                  if player.multiplierCounter >= 10 then
                    player.multiplierCounter = 0
                    if player.multiplier < 10 then player.multiplier = player.multiplier + 1 end
                  end
                  player.kills = player.kills + 1
                end
              end
            end
          end
        end
      end

      for i, b in pairs(bullet) do
        if b.who == "enemy" and player.dying == false then
          if player.hasShield then
            if distanceFrom(player.x, player.y, b.x, b.y) < player.shieldRad and b.dead == false then
              b.dead = true
  --            spawnExplosion(b.x,b.y,1)
            end
          else
            if distanceFrom(player.x, player.y, b.x, b.y) < 16 and b.dead == false then
              b.dead = true
              spawnExplosion(player.x,player.y,1)
              damagePlayer(10)
            end
          end
        end

        if b.target <= 0 then
          b.life = b.life - dt
          b.x = b.x + (_c(b.angle * pi / 180) * (b.speed * dt)) * 1
          b.y = b.y + (_s(b.angle * pi / 180) * (b.speed * dt)) * 1
          if b.life < 0 then b.dead = true end
        else
          if b.who == "player" then
            if enemy[b.target] then
              b.angle = -findAngle(b.x, b.y, enemy[b.target].x, enemy[b.target].y)
            end
            b.x = b.x + (_c(b.angle * pi / 180) * (b.speed * dt)) * 1
            b.y = b.y + (_s(b.angle * pi / 180) * (b.speed * dt)) * 1
          end
        end
        if b.x < level.offset-10 or b.x > level.offset+810 or b.y < -10 or b.y > 510 then
          b.dead = true
        end
      end

      if player.energy <= 0 and player.dying == false then
        player.fallSpeed = level.speed
        player.dying = true
      end

      if player.dying and player.dead == false then
        player.fallSpeed = player.fallSpeed + 50 * dt
        player.deathFrame = player.deathFrame - dt
        if player.deathFrame <= 0 then
          player.deathFrame = _r(15,25)/100
          spawnExplosion(player.x, player.y, 1)
        end
        player.x = player.x + (player.fallSpeed*1.6) * dt
        player.y = player.y + player.fallSpeed * dt
      end

      updateExplosions(dt)
    end
  end
end

function love.draw(dt)
  if inMenu then
    gr.setLineWidth(lw)
    drawWord("SHMUP!",50,10,64,96,{255,255,255})
    for i, m in pairs(menu.item) do
      if m.x < m.xt then m.x = m.x + m.s*dt end
      local c = {255,255,255}
      if i == menu.selection then
        if menu.blink == true then
          if _s(time*50) > 0 then
            c = {0,0,0,0}
          else
            c = {255,255,0}
          end
        else
          c = {255,255,0}
        end
      end
      drawWord(m.title,50+menu.item[i].x,100 + (96*i),32,64,c)
    end
  else
    gr.setLineWidth(lw)
    local fps = love.timer.getFPS()

    local xlb = _f((level.offset)/32)
    local xub = xlb + 26
    if xlb < 0 then xlb = 0 end
    if xub > level.length then xub = level.length end

    local xlb2 = _f(xlb*.5)-4
    local xub2 = xlb2 + 52
    if xlb2 < 0 then xlb2 = 0 end
    if xub2 > level.length * 2 then xub2 = level.length * 2 end

    -- DRAW SUN/MOON
    local sx, sy = (level.sunmoon.x-level.offset/20)*screenScale, (level.sunmoon.y)*screenScale
    gr.setColor(144,148,62)
    gr.circle("fill", sx, sy, 48*screenScale, 100)
    gr.circle("line", sx, sy, 48*screenScale, 100)

    for x=xlb2,xub2 do
      if x > 0 then
        if level.para[x].cloud > 0 then
          gr.setColor(unpack(level.para[x].cloudC))
          gr.circle("line", ((x*32)-level.paraOffset)*screenScale,0, level.para[x].cloud*screenScale, 70)
          gr.circle("fill", ((x*32)-level.paraOffset)*screenScale,0, level.para[x].cloud*screenScale, 70)
        end

        if level.para[x].mountain > 0 then
          gr.setColor(unpack(level.para[x].mountainC))
          gr.triangle("line",
            ((x*32)-level.paraOffset-level.para[x].mountain*.66)*screenScale, 450*screenScale,
            ((x*32)-level.paraOffset+level.para[x].mountain*.66)*screenScale, 450*screenScale,
            ((x*32)-level.paraOffset)*screenScale, (450-level.para[x].mountain)*screenScale
          )
          gr.triangle("fill",
            ((x*32)-level.paraOffset-level.para[x].mountain*.66)*screenScale, 450*screenScale,
            ((x*32)-level.paraOffset+level.para[x].mountain*.66)*screenScale, 450*screenScale,
            ((x*32)-level.paraOffset)*screenScale, (450-level.para[x].mountain)*screenScale
          )
        end
      end
    end

    -- UPDATE TILE CHANGE
    if xlb ~= oxlb then
      loadEnemies(xub)

      if level.map[xlb].checkpoint then
        player.checkpoint.x = xlb
        player.checkpoint.speed = level.speed
        player.checkpoint.sunmoonY = level.sunmoon.y
        insertMessage("CHECKPOINT! " .. xlb)
      end
    end

    -- DRAW POWERUPS
    for i, p in pairs(powerup) do
      if p.dead == false then
        gr.setColor(255,255,255)
        gr.circle("line", (p.x-level.offset)*screenScale, p.y*screenScale, 10*screenScale, 40)

        if distanceFrom(player.x, player.y, p.x, p.y) <= 26 then
          powerup[i].give()
          p.dead = true
        end
      end
    end

    -- DRAW BULLETS!
    for i, b in pairs(bullet) do
      if b.dead == false then
        gr.setColor(255,0,0)
        --gr.circle("line", (b.x-level.offset)*screenScale, b.y*screenScale, 3*screenScale)
        --gr.setLineWidth(2*screenScale)
        local bx2 = b.x + (_c((b.angle-180) * pi / 180) * (25)) * 1
        local by2 = b.y + (_s((b.angle-180) * pi / 180) * (25)) * 1

        gr.line((b.x-level.offset)*screenScale, b.y*screenScale, (bx2-level.offset)*screenScale, by2*screenScale)
      end
    end

    -- DRAW ENEMIES!
    local c1, c2
    for i, e in pairs(enemy) do
      if e.type >= 100 then
        e.draw(i)
      else
        if e.dead == false then
          if e.type == 1 then
            c1 = {255,0,0}
            c2 = {255,0,0,255/2}
          elseif e.type == 2 then
            c1 = {0,0,255}
            c2 = {0,0,255,255/2}
          end
          gr.setColor(unpack(c2))
          gr.circle("fill", (e.x-level.offset)*screenScale, e.y*screenScale, e.radius*screenScale, 50)
          gr.setColor(unpack(c1))
          gr.circle("line", (e.x-level.offset)*screenScale, e.y*screenScale, e.radius*screenScale, 50)
          if e.powerup > 1 then
            drawWord("PU", (e.x-level.offset)*screenScale, e.y*screenScale, 16,16, {255,255,255})
          end
        end
      end
    end

    if player.shieldRad > 0 then
      gr.setColor(0,0,255,255/10)
      gr.circle("fill", (player.x-level.offset)*screenScale, player.y*screenScale, player.shieldRad+(_s(time*5)*3), 40)
      gr.setColor(0,0,255)
      gr.circle("line", (player.x-level.offset)*screenScale, player.y*screenScale, player.shieldRad+(_s(time*5)*3), 40)
    end

    if player.dead == false then
      gr.setColor(0,255/2,0)
      gr.circle("fill", (player.x-level.offset)*screenScale, player.y*screenScale, 16*screenScale, 20)
      gr.setColor(0,255,0)
      gr.circle("line", (player.x-level.offset)*screenScale, player.y*screenScale, 16*screenScale, 20)
    end

    drawExplosions()

    if level.liquid > 0 then
      local bob = (500-level.liquidHeight) + (4 * _s(time))
      local bob2 = (500-level.liquidHeight) + (4 * _c(time))
      gr.setColor(unpack(color[level.liquid][1]))
      gr.polygon("fill",
        0, bob*screenScale,
        800*screenScale, bob2*screenScale,
        800*screenScale, 500*screenScale,
        0, 500*screenScale
      )
      gr.setColor(unpack(color[level.liquid][2]))
      gr.line(0,bob*screenScale, 800*screenScale,bob2*screenScale)
    end

    for x=xlb,xub do
      if x > 0 and x < level.length then
        if level.map[x].fDeco then
          gr.circle("fill", ((x*32)-level.offset)*screenScale+16, level.map[x-1].floor*screenScale+.5-10, 10)
          gr.setColor(95/2,56/2,21/2)
          gr.rectangle("fill", (
            (x*32)-level.offset)*screenScale+16, level.map[x-1].floor*screenScale+.5-10, 4, 10
          )
        end

        --CEILING
        if level.map[x].ceilType > 0 then
          gr.setColor(unpack(color[level.map[x].ceilType][1]))
          gr.polygon("fill",
            (((x-1)*32)-level.offset)*screenScale+.5, 0,
            ((x*32)-level.offset)*screenScale+.5, 0,
            ((x*32)-level.offset)*screenScale+.5, level.map[x].ceil*screenScale+.5,
            (((x-1)*32)-level.offset)*screenScale+.5, level.map[x-1].ceil*screenScale+.5
          )

          gr.setColor(unpack(color[level.map[x].ceilType][2]))
          gr.line(
            (((x-1)*32)-level.offset)*screenScale+.5,
            level.map[x-1].ceil*screenScale+.5,
            ((x*32)-level.offset)*screenScale+.5,
            level.map[x].ceil*screenScale+.5
          )
        end

        if (level.map[x-1].floor > 0 or level.map[x].floor > 0) and level.map[x].floorType > 0 then
          --FLOOR
          gr.setColor(unpack(color[level.map[x].floorType][1]))
          gr.polygon("fill",
            (((x-1)*32)-level.offset)*screenScale+.5, 500*screenScale,
            ((x*32)-level.offset)*screenScale+.5, 500*screenScale,
            ((x*32)-level.offset)*screenScale+.5, (500-level.map[x].floor)*screenScale+.5,
            (((x-1)*32)-level.offset)*screenScale+.5, (500-level.map[x-1].floor)*screenScale+.5
          )

          gr.setColor(unpack(color[level.map[x].floorType][2]))
          gr.line(
            (((x-1)*32)-level.offset)*screenScale+.5,
            (500-level.map[x-1].floor)*screenScale+.5,
            ((x*32)-level.offset)*screenScale+.5,
            (500-level.map[x].floor)*screenScale+.5
          )
        end
      end
    end

    if debugVar then
      drawWord("FPS:      "..fps,10,10,8,12,{255,255,255})
      drawWord("BULLETS:  "..#bullet,10,30,8,12,{255,255,255})
      drawWord("ENEMIES:  "..#enemy,10,50,8,12,{255,255,255})
      drawWord("EXPLOSION:"..#explosion,10,70,8,12,{255,255,255})
      drawWord("L SPEED:  "..level.speed,10,90,8,12,{255,255,255})
      drawWord("TILE X:   "..player.tileX,10,110,8,12,{255,255,255})
      drawWord("CHECK:    "..player.checkpoint.x,10,130,8,12,{255,255,255})
      drawWord("OFFSET:   "..level.offset,10,150,8,12,{255,255,255})
      drawWord("PARAOFF:  "..level.paraOffset,10,170,8,12,{255,255,255})
      drawWord("XLB/XUB:  "..xlb.."/"..xub,10,190,8,12,{255,255,255})
      drawWord("XLB2/XUB2:"..xlb2.."/"..xub2,10,210,8,12,{255,255,255})
      drawWord("SUN/MOON: "..level.sunmoon.y,10,230,8,12,{255,255,255})

      drawWord("BULLETS:",700,10,5,6,{255,255,255})
      for i, b in pairs(bullet) do
        if b.who == "player" then c = {255,0,0} else c = {255,255,0} end
        drawWord(_f(b.x) .. "," .. _f(b.y),700,20+(i*8),5,6,c)
      end
    end

    drawWord("KILLS:"..formatNumber(player.kills),20,screenH/screenScale-30,16,20,{255,255,255})

    if level.gameOver then
      if _s(time*20) > 0 then
        drawWord("GAME OVER!",150,130,48,64,{255,255,255})
      end
    end

    if level.readyMessage > 0 then
      if _s(time*20) > 0 then
        drawWord("READY!!!",200,130,48,64,{255,255,255})
      end
    end

    local t = ""
    local c
    if player.lives > 3 then
      t = ">x"..player.lives
    else
      for i=1,player.lives do
        t = t .. ">"
      end
    end
    drawWord(t,590,screenH/screenScale-30,16,20,{255,255,255})

    t = "ENERGY"
    c = {255,255,255}
    if player.energy <= 25 then
      t = "DANGER!"
      if _s(time*50) > 0 then c = {255,0,0} else c = {255,255,255} end
    end
    drawWord(t,680, (screenH/screenScale-38),5,6,c)

    local h = 20
    if player.hasShield then
      h = 10

      gr.setColor(0,0,255)
      gr.rectangle("fill", 680*screenScale+.5, (screenH/screenScale-20)*screenScale+.5, player.shieldTime*10*screenScale, (h-1)*screenScale)
    end

    if player.hitFlash > 0 and _s(time*50) > 0 then
      gr.setColor(255,0,0)
    else
      if player.hasShield then
        gr.setColor(255,255/2,0,255/2)
      else
        gr.setColor(255,255,255,255/2)
      end
    end

    gr.rectangle("fill", 680*screenScale+.5, (screenH/screenScale-30)*screenScale+.5, player.energy*screenScale, (h-1)*screenScale)
    if player.energy <= 25 then
      t = "DANGER!"
      if _s(time*50) > 0 then c = {255,0,0} else c = {255,255,255} end
    end
    gr.setColor(unpack(c))
    gr.rectangle("line", 680*screenScale+.5, (screenH/screenScale-30)*screenScale+.5, 100*screenScale, h*screenScale)
    if player.hasShield then
      gr.rectangle("line", 680*screenScale+.5, (screenH/screenScale-20)*screenScale+.5, 100*screenScale, h*screenScale)
    end

    if player.multiplier > 1 then drawWord("MULTIPLIER:"..player.multiplier,340,screenH/screenScale-38,5,6,{255,255,255}) end
    drawWord(formatNumber(player.score),340,screenH/screenScale-30,16,20,{255,255,255})


    oxlb = xlb
    oxlb2 = xlb2

    local s = 1+_s(time)

    if paused then
      doPauseColor()
      drawWord("PAUSED", 295,100,32,40,pauseColor)
    end

    if level.rain then
      if paused == false then level.timeSinceDrop = level.timeSinceDrop - dt end
      for i, r in pairs(rainDrop) do
        if paused == false then
          r.x = r.x - level.speed * dt
          r.y = r.y + 512 * dt * r.s
        end
        if r.y > 500 then table.remove(rainDrop, i) end
        local x, y, x2, y2
        x = r.x*screenScale - level.offset
        y = r.y*screenScale
        x2 = (r.x-4*r.s)*screenScale - level.offset
        y2 = (r.y+32*r.s)*screenScale

        gr.setColor(255,255,255,100)
        gr.line(x,y,x2,y2)
      end
      if level.timeSinceDrop <= 0 then
        level.timeSinceDrop = .05
        table.insert(rainDrop, {x = _r(level.offset,level.offset+800)+32, y = 0, s = _r(1,2)/2})
      end
    end

    drawMessages(dt)
    removeJunk()
  end
end

function love.keypressed(k)
  if inMenu then
    if k == "return" or k == "enter" or k == " " then
      menu.item[menu.selection].go()
    elseif k == "up" then
      if menu.selection > 1 then menu.selection = menu.selection - 1 end
    elseif k == "down" then
      if menu.selection < #menu.item then menu.selection = menu.selection + 1 end
    end
  else
    if player.dying == false then
      if k == "p" then
        if paused then paused = false else paused = true end
      elseif k == "lmeta" then
        giveSheild(10)
      elseif k == "g" then
        genocide()
      elseif k == "k" then
        player.energy = 0
      end

      if k == "escape" then setupMenu() end

      if k == "`" then debugVar = not debugVar end

      if k == "1" or k == "2" or k == "3" then
        switchPrimaryWeapon(tonumber(k))
      end
    end
  end
end

function givePowerup(i)

end

function removeJunk()
  for i=#enemy,1,-1 do
    if enemy[i].dead == true then
      table.remove(enemy, i)
    end
  end

  for i=#explosion,1,-1 do
    if explosion[i].dead == true then
      table.remove(explosion, i)
    end
  end

  for i=#bullet,1,-1 do
    if bullet[i].dead == true then
      table.remove(bullet, i)
    end
  end

  for i=#powerup,1,-1 do
    if powerup[i].powerup == true then
      table.remove(powerup, i)
    end
  end
end

function updateBG(dt)
  if bg.r < bg.rt then bg.r = bg.r + dt * bg.speed if bg.r > bg.rt then bg.r = bg.rt end end
  if bg.g < bg.gt then bg.g = bg.g + dt * bg.speed if bg.g > bg.gt then bg.g = bg.gt end end
  if bg.b < bg.bt then bg.b = bg.b + dt * bg.speed if bg.b > bg.bt then bg.b = bg.bt end end

  if bg.r > bg.rt then bg.r = bg.r - dt * bg.speed if bg.r < bg.rt then bg.r = bg.rt end end
  if bg.g > bg.gt then bg.g = bg.g - dt * bg.speed if bg.g < bg.gt then bg.g = bg.gt end end
  if bg.b > bg.bt then bg.b = bg.b - dt * bg.speed if bg.b < bg.bt then bg.b = bg.bt end end

  gr.setBackgroundColor(bg.r, bg.g, bg.b)
end

function setBG(r,g,b)
  bg.rt = r
  bg.gt = g
  bg.bt = b
end

function setTimeout(go, t)
  table.insert(timeout, { go = go, time = t})
end

function updateTimeout(dt)
  for i, t in pairs(timeout) do
    t.time = t.time - dt
    if t.time <= 0 then
      t.go()
      table.remove(timeout, i)
    end
  end
end

function unlockAchievement(n)
  if achievements[n].unlocked == false then
    achievements[n].unlocked = true
    print(achievements[n].desc)
    insertMessage(achievements[n].desc)
  end
end

function setupMenu()
  setBG(0,0,0)
  local getSoundValue = function()
    local boo = {"OFF", "ON"}
    return boo[(soundOn and 1 or 0)+1]
  end
	menu = {
	  selection = 1,
	  item = {},
	  blink = false
	}
	menu.item[1] = {
	  title = "PLAY",
	  x = 0, xt = 0, s = 800,
	  go = function()
	    menu.item[2].xt = 800
	    menu.item[3].xt = 800
	    --menu.item[1].xt = 800
	    menu.item[2].s = 800
	    menu.item[3].s = 1600
	    playSound("explode_1")
	    menu.blink = true
	    setTimeout(function() setupGame() end, 1)
	    setTimeout(function() menu.blink = false end, .75)
	  end
	}
	menu.item[2] = {
	  title = "SOUND:"..getSoundValue(),
	  x = 0, xt = 0, s = 800,
	  go = function()
	    menu.blink = true
      if soundOn == true then soundOn = false else soundOn = true end
      menu.item[2].title = "SOUND:"..getSoundValue()
	    playSound("explode_1")
	    setTimeout(function() menu.blink = false end, .75)
	  end
	}
	menu.item[3] = {
	  title = "QUIT",
	  x = 0, xt = 0, s = 800,
	  go = function()
	    menu.item[1].xt = 800
	    menu.item[2].xt = 800
	    --menu.item[3].xt = 800
	    menu.item[1].s = 800
	    menu.item[2].s = 1600
	    playSound("explode_1")
	    menu.blink = true
	    setTimeout(function() love.audio.stop() love.event.push("q") end, 1)
	    setTimeout(function() menu.blink = false end, .75)
	  end
	}

	inMenu = true
end

function setupGame()
  loadEnemyTypes()

	primaryWeapon = {}
	primaryWeapon[1] = {
	  timing = .25, sound = "player_fire_1", fire = function()
      spawnBullet(player.x, player.y, 1, 0, 512, "player")
	  end
	}
	primaryWeapon[2] = {
	  timing = .125, sound = "player_fire_1", fire = function()
      spawnBullet(player.x, player.y-12, 1, 0, 768, "player")
      spawnBullet(player.x, player.y+12, 1, 0, 768, "player")
	  end
	}
	primaryWeapon[3] = {
	  timing = 1, sound = "player_fire_1", fire = function()
      spawnBullet(player.x, player.y, 3, 0, 256, "player", 1)
	  end
	}

	powerupList = {}
	powerupList[1] = {
	  give = function()
      playSound("powerup_1")
      switchPrimaryWeapon(2)
	  end
	}

  water, lava, slime = 11, 12, 13
  dirt, grass, stone, none = 1, 2, 3, 0

  player = {
    x = 50, y = 250, tileX = 0, xs = 256, ys = 256, invTime = .5, inv = 0,
    energy = 100, energyMax = 100, hasShield = false, shieldTime = 0, shieldRad = 0, hitFlash = 0, drifting = 0,
    lives = 3, kills = 0, timeShot = 0, timeBetween = .2, multiplier = 1, multiplierCounter = 0, score = 0,
    primary = 1, secondary = 0, checkpoint = {x = 0, speed = 0, sunmoonY = 0}, dying = false, dead = false, fallSpeed = 0, deathFrame = 0
  }

	level = {
	  offset = 0, paraOffset = 0, speed = 0, speedTo = 64, map = {}, para = {},
	  liquid = water, liquidHeight = 56, rain = true, timeSinceDrop = 0, sunmoon = {x = 750, y = 450},
	  scrollPause = 1, background = {62,110,124}, readyMessage = 0, gameOver = false, length = 50
	}

  color = {}
  color[dirt] = { {95/2,56/2,21/2}, {95,56,21} }
  color[grass] = { {0,255/2,0}, {0,255,0} }

  color[water] = { {0,0,255/2}, {0,0,255} }
  color[lava] = { {255,255/2,0}, {255,255,0} }
  color[slime] = { {0,255/2,0}, {0,255,0} }

  for x=0,level.length*2-1 do
	  level.para[x] = {}
	  local c = 0
	  local c1 = level.background[1]+(255-level.background[1])*.33
	  local c2 = level.background[2]+(255-level.background[2])*.33
	  local c3 = level.background[3]+(255-level.background[3])*.33
	  if x/3 == _f(x/3) then c = _r(2,4) end
    level.para[x].cloud = c * 32
    level.para[x].cloudC = {c1,c2,c3}

	  c1 = level.background[1]-(level.background[1])*.33
	  c2 = level.background[2]-(level.background[2])*.33
	  c3 = level.background[3]-(level.background[3])*.33
	  if _r(1,2) == 1 and x/3 == _f(x/3) then c = _r(2,6) end
    level.para[x].mountain = c * 32
    level.para[x].mountainC = {c1,c2,c3}
  end

  local pu
	for x=0,level.length-1 do
	  level.map[x] = {}
	  for y=0,15-1 do
	    local e
	    if _r(1,20) == 1 and x > 24 and y > 2 and y < 13 then
	      e = _r(1,2)
	    else
	      e = 0
	    end
	    pu = 0
	    if e > 0 then
	      if _r(1,4) == 1 then
	        pu = 1
	        print("PU", x, y)
	      end
	    end
      level.map[x][y] = { tile = 0, enemy = e, powerup = pu }
	  end

	  level.map[x].ceil = -2--_r(24,40)
	  level.map[x].ceilType = none
	  level.map[x].floor = 40+_r(24,40)
	  level.map[x].floorType = grass
	  level.map[x].enemies = {}
    if _r(0,3) == 0 then
  	  --level.map[x].fDeco = _r(1,2)
  	end
	end

  local wl = _r(10,20)
  local wl2 = wl+_r(15,25)
  for i=wl,wl2 do
    level.map[i].floor = 0
  end

  local j = 1
  for i=wl-4,wl do
    level.map[i].floor = level.map[i].floor / j
    j = j + 1
  end

  local j = 5
  for i=wl2,wl2+4 do
    level.map[i].floor = level.map[i].floor / j
    j = j - 1
  end

  for i=1,5 do
    level.map[_r(30,level.length*.8)].setSpeed = _r(32,256)
  end

  --for i=1,level.length/100-1 do
    --level.map[(i*100)].checkpoint = true
  --end

  enemy = {}
  bullet = {}
  explosion = {}
  rainDrop = {}
  powerup = {}

  messages = {}

	fsmall = gr.newFont(love._vera_ttf, 12*screenScale)
	fbig = gr.newFont(love._vera_ttf, 20*screenScale)

  paused = false

  switchPrimaryWeapon(1)

  level.readyMessage = 1

  setBG(unpack(level.background))

  if soundOn then
    --love.audio.play(bgm)
  end

  inMenu = false
end

function endGame()
  level.winnerScreen = true
  setTimeout(function()
    setupMenu()
  end, 4)
end

function switchPrimaryWeapon(w)
  player.primary = w
  player.timeBetween = primaryWeapon[w].timing
  player.timeShot = 0
end

function playerFireWeapon()
  primaryWeapon[player.primary].fire()
  playSound(primaryWeapon[player.primary].sound)
end

function doPauseColor()
  pauseColor[1] = _r(100,255)
  pauseColor[2] = _r(100,255)
  pauseColor[3] = _r(100,255)
end

function killPlayer()
  player.lives = player.lives - 1
  if player.lives > 0 then
    setTimeout(function() resetPlayer() end, 2)
  else
    level.gameOver = true
    setTimeout(function() resetPlayer() end, 5)
  end
end

function resetPlayer()
  if player.lives > 0 then
    player.dying = false
    player.dead = false
    player.fallSpeed = 0
    explosion = {}
    bullet = {}
    enemy = {}
    if player.checkpoint.x == 0 then
      level.offset = 0
      level.speed = 64
      level.sunmoon.y = 450
    else
      level.offset = (player.checkpoint.x+1) * 32
      level.speed = player.checkpoint.speed
      level.sunmoon.y = player.checkpoint.sunmoonY
    end
    level.speedTo = level.speed
    player.x = level.offset + 40
    player.y = 250
    level.scrollPause = 3
    player.energy = player.energyMax
    player.multiplier = 1
    level.readyMessage = 1
    giveSheild(4)
  else
    setupMenu()
  end
end

function giveSheild(t)
  player.hasShield = true
  player.shieldTime = t
end

function damagePlayer(d)
  if player.inv <= 0 then
    if player.hasShield then

    else
      player.energy = player.energy - d
      player.hitFlash = .5
      player.inv = player.invTime
    end
  end
end

function loadAlphabet()
  letter = {}
  letter["A"] = {
    { { 0,1 }, { .5,0 } },
    { { .5,0 }, { 1,1 } },
    { { .25,.5 }, { .75,.5 } }
  }
  letter["B"] = {
    { { 0,0 }, { 0,1 } },
    { { 0,0 }, { .75,0 } },
    { { 0,.5 }, { 1,.5 } },
    { { 0,1 }, { 1,1 } },
    { { .75,0 }, { .75,.5 } },
    { { 1,.5 }, { 1,1 } }
  }
  letter["C"] = {
    { { 0,0 }, { 1,0 } },
    { { 0,0 }, { 0,1 } },
    { { 0,1 }, { 1,1 } }
  }
  letter["D"] = {
    { { 0,0 }, { 0,1 } },
    { { 0,0 }, { .75,0 } },
    { { .75,0 }, { 1,.25 } },
    { { 1,.25 }, { 1,.75 } },
    { { 1,.75 }, { .75,1 } },
    { { .75,1 }, { 0,1 } }
  }
  letter["E"] = {
    { { 0,0 }, { 1,0 } },
    { { 0,0 }, { 0,1 } },
    { { 0,1 }, { 1,1 } },
    { { 0,.5 }, { .75,.5 } }
  }
  letter["F"] = {
    { { 0,0 }, { 1,0 } },
    { { 0,0 }, { 0,1 } },
    { { 0,.5 }, { .75,.5 } }
  }
  letter["G"] = {
    { { 0,0 }, { 1,0 } },
    { { 0,0 }, { 0,1 } },
    { { 0,1 }, { 1,1 } },
    { { 1,1 }, { 1,.5 } },
    { { 1,.5 }, { .5,.5 } }
  }
  letter["H"] = {
    { { 0,1 }, { 0,0 } },
    { { 1,0 }, { 1,1 } },
    { { 0,.5 }, { 1,.5 } }
  }
  letter["I"] = {
    { { .5,0 }, { .5,1 } }
  }
  letter["J"] = {
    { { 1,0 }, { 1,1 } },
    { { 1,1 }, { 0,1 } },
    { { 0,1 }, { 0,.75 } }
  }
  letter["K"] = {
    { { 0,0 }, { 0,1 } },
    { { 0,.5 }, { 1,0 } },
    { { 0,.5 }, { 1,1 } }
  }
  letter["L"] = {
    { { 0,0 }, { 0,1 } },
    { { 0,1 }, { 1,1 } }
  }
  letter["M"] = {
    { { 0,0 }, { 0,1 } },
    { { 0,0 }, { .5,1 } },
    { { .5,1 }, { 1,0 } },
    { { 1,1 }, { 1,0 } }
  }
  letter["N"] = {
    { { 0,0 }, { 0,1 } },
    { { 0,0 }, { 1,1 } },
    { { 1,1 }, { 1,0 } }
  }
  letter["O"] = {
    { { 0,0 }, { 1,0 } },
    { { 0,0 }, { 0,1 } },
    { { 0,1 }, { 1,1 } },
    { { 1,0 }, { 1,1 } }
  }
  letter["P"] = {
    { { 0,0 }, { 1,0 } },
    { { 0,0 }, { 0,1 } },
    { { 0,.5 }, { 1,.5 } },
    { { 1,0 }, { 1,.5 } }
  }
  letter["Q"] = {
    { { 0,0 }, { 1,0 } },
    { { 0,0 }, { 0,1 } },
    { { 0,1 }, { 1,1 } },
    { { 1,0 }, { 1,1 } },
    { { .75,.75 }, { 1.125,1.125 } }
  }
  letter["R"] = {
    { { 0,0 }, { 1,0 } },
    { { 0,0 }, { 0,1 } },
    { { 0,.5 }, { 1,.5 } },
    { { 1,0 }, { 1,.5 } },
    { { .5,.5 }, { 1,1 } }
  }
  letter["S"] = {
    { { 1,0 }, { 0,0 } },
    { { 0,0 }, { 0,.5 } },
    { { 0,.5 }, { 1,.5 } },
    { { 1,.5 }, { 1,1 } },
    { { 1,1 }, { 0,1 } }
  }
  letter["T"] = {
    { { .5,0 }, { .5,1 } },
    { { 0,0 }, { 1,0 } }
  }
  letter["U"] = {
    { { 0,0 }, { 0,1 } },
    { { 0,1 }, { 1,1 } },
    { { 1,1 }, { 1,0 } }
  }
  letter["V"] = {
    { { 0,0 }, { .5,1 } },
    { { .5,1 }, { 1,0 } }
  }
  letter["W"] = {
    { { 0,0 }, { .25,1 } },
    { { .25,1 }, { .5,0 } },
    { { .5,0 }, { .75,1 } },
    { { .75,1 }, { 1,0 } }
  }
  letter["X"] = {
    { { 0,0 }, { 1,1 } },
    { { 1,0 }, { 0,1 } }
  }
  letter["Y"] = {
    { { 0,0 }, { .5,.5 } },
    { { .5,.5 }, { 1,0 } },
    { { .5,.5 }, { .5,1 } }
  }
  letter["Z"] = {
    { { 0,0 }, { 1,0 } },
    { { 1,0 }, { 0,1 } },
    { { 0,1 }, { 1,1 } }
  }

  letter["x"] = {
    { { 0,.25 }, { 1,1 } },
    { { 1,.25 }, { 0,1 } }
  }

  letter["!"] = {
    { { .5,0 }, { .5,.7 } },
    { { .5,.85 }, { .5,1 } }
  }
  letter["."] = {
    { { .5,.85 }, { .5,1 } }
  }
  letter[":"] = {
    { { .15,.85 }, { .15,1 } },
    { { .15,0 }, { .15,.15 } }
  }
  letter[","] = {
    { { .5,.85 }, { .5,1 } },
    { { .5,1 }, { .35,1.12 } }
  }
  letter["$"] = {
    { { 1,0 }, { 0,0 } },
    { { 0,0 }, { 0,.5 } },
    { { 0,.5 }, { 1,.5 } },
    { { 1,.5 }, { 1,1 } },
    { { 1,1 }, { 0,1 } },
    { { .5,-.2 }, { .5,1.2 } }
  }

  letter["0"] = {
    { { 0,0 }, { 1,0 } },
    { { 1,0 }, { 1,1 } },
    { { 1,1 }, { 0,1 } },
    { { 0,1 }, { 0,0 } },
    { { 1,0 }, { 0,1 } }
  }
  letter["1"] = {
    { { .5,0 }, { .5,1 } },
    { { .5,0 }, { .25,.25 } },
    { { .25,1 }, { .75,1 } }
  }
  letter["2"] = {
    { { 0,1 }, { 1,1 } },
    { { 0,1 }, { 0,.75 } },
    { { 0,.75 }, { 1,.25 } },
    { { 1,.25 }, { 1,0 } },
    { { 1,0 }, { 0,0 } },
    { { 0,0 }, { 0,.25 } }
  }
  letter["3"] = {
    { { 0,0 }, { 1,0 } },
    { { .25,.5 }, { 1,.5 } },
    { { 0,1 }, { 1,1 } },
    { { 1,0 }, { 1,1 } }
  }
  letter["4"] = {
    { { .75,0 }, { .75,1 } },
    { { 0,.5 }, { 1,.5 } },
    { { .125,0 }, { 0,.5 } }
  }
  letter["5"] = {
    { { 1,0 }, { 0,0 } },
    { { 0,0 }, { 0,.5 } },
    { { 0,.5 }, { 1,.5 } },
    { { 1,.5 }, { 1,1 } },
    { { 1,1 }, { 0,1 } }
  }
  letter["6"] = {
    { { 1,0 }, { 0,0 } },
    { { 0,0 }, { 0,1 } },
    { { 0,.5 }, { 1,.5 } },
    { { 1,.5 }, { 1,1 } },
    { { 1,1 }, { 0,1 } }
  }
  letter["7"] = {
    { { 0,0 }, { 1,0 } },
    { { 1,0 }, { .5,1 } }
  }
  letter["8"] = {
    { { 1,0 }, { 0,0 } },
    { { 0,0 }, { 0,1 } },
    { { 0,.5 }, { 1,.5 } },
    { { 1,0 }, { 1,1 } },
    { { 1,1 }, { 0,1 } }
  }
  letter["9"] = {
    { { 1,0 }, { 0,0 } },
    { { 0,0 }, { 0,.5 } },
    { { 0,.5 }, { 1,.5 } },
    { { 1,0 }, { 1,1 } },
    { { 1,1 }, { 0,1 } }
  }

  letter[">"] = {
    { { 0,0 }, { 1,.5 } },
    { { 1,.5 }, { 0,1 } },
    { { 0,1 }, { 0,0 } }
  }

end

function drawWord(t,x,y,sx,sy,c)
  t = tostring(t)
  local l
  local xx = 0
  local spa = (sx/10)
  if spa < 2 then spa = 2 end
  for i=1,#t do
    l = string.sub(t,i,i)
    drawLetter(l,x+xx,y,sx,sy,c)
    xx = xx + _f(sx + spa)
  end
end

function drawLetter(l,x,y,sx,sy,c)
  if letter[l] then
    local s1 = screenScale*sx
    local s2 = screenScale*sy
    x = x * screenScale
    y = y * screenScale

    gr.setColor(unpack(c))
    for a=1,#letter[l] do
      gr.line(x+(letter[l][a][1][1]*s1+.5), y+(letter[l][a][1][2]*s2+.5), x+(letter[l][a][2][1]*s1+.5), y+(letter[l][a][2][2]*s2+.5))
    end
  end
end

function genocide()
  for i=1,45 do
    setTimeout(function()
      spawnBullet(player.x, player.y, 1, (i*2)-45, 512, "player")
    end, i*.001)
  end
end

function loadEnemies(x)
  if x < level.length then
    if level.scrollPause <= 0 then
      for y=0,15-1 do
        if level.map[x][y].enemy > 0 then
          if level.map[x][y].enemy == 1 then
            en = 1
          else
            en = 2
          end
          enemy[#enemy+1] = { x=x*32, y=y*32, points = 100, energy = en, type = level.map[x][y].enemy, tsb = 0, radius = 16, dead = false, powerup = level.map[x][y].powerup }
        end
      end
      if level.map[x].setSpeed then
        level.speedTo = level.map[x].setSpeed
      end
    end
  else
    level.speed = 0
    level.speedTo = 0
    enemyType[100].create(level.length*32+128,250)
--    level.spawnBoss(101,level.length*32+128,100)
  end
end

function spawnExplosion(x,y,t)
  if t == 1 then

  elseif t == 2 then

  end
  local ra = _r(0,359)
  for i=1,5 do
    explosion[#explosion+1] = { x = x, y = y, type = t, life = 32, angle = ((360/5)*i), rot = ra, dead = false }
  end
end

function updateExplosions(dt)
  for i, e in pairs(explosion) do
    e.rot = e.rot + 128*dt
    e.x = e.x + (_c((e.angle-180+e.rot) * pi / 180) * (64)) * dt
    e.y = e.y + (_s((e.angle-180+e.rot) * pi / 180) * (64)) * dt
    e.life = e.life - 64 * dt
    e.y = e.y - 64 * dt
    if e.life < 0 then e.dead = true end
  end
end

function drawExplosions()
  for i, e in pairs(explosion) do
    if e.dead == false then
      if e.type == 1 then
        gr.setColor(255,(255-(7*e.life)),0,(255-(3*e.life))/2)
      elseif e.type == 2 then
        gr.setColor((255-(7*e.life)),(255-(7*e.life)),255,(255-(3*e.life))/2)
      end

      gr.circle("fill", (e.x-level.offset)*screenScale, e.y*screenScale, e.life*screenScale*.75, 32)
    end
  end
end

function spawnBullet(x,y,t,a,s,w,g)
  if t == 1 then
    l = 10
    p = 1
  elseif t == 2 then
    l = 10
    p = 1
  elseif t == 3 then
    l = 10
    p = 10
  else
    l = 10
    p = 1
  end
  if g == nil then
    g = -1
  end

  bullet[#bullet+1] = { x = x, y = y, type = t, speed = s, angle = a, life = l, who = w, dead = false, target = g, power = p }
end

function insertMessage(m)
  table.insert(messages, {text = m, time = 3})
end

function drawMessages(dt)
  if #messages > 0 then
    drawWord(messages[1].text,260,60,12,18,{255,255,255})
    messages[1].time = messages[1].time - dt
    if messages[1].time <= 0 then
      table.remove(messages, 1)
    end
  end
end

function degToRad(d) return d * pi / 180 end

function sort(T) table.sort(T, function(a, b) return a.y < b.y end ) end

function formatNumber(number) return (string.format("%d", number):reverse():gsub( "(%d%d%d)" , "%1," ):reverse():gsub("^(-?),","%1")) end

function overlap(box1x, box1y, box1w, box1h, box2x, box2y, box2w, box2h) if box1x > box2x + box2w - 1 or box1y > box2y + box2h - 1 or box2x > box1x + box1w - 1 or box2y > box1y + box1h - 1 then return false else return true end end

function playSound(s)
  if soundOn then soundmanager:play(sfx[s]) end
end

function distanceFrom(x1,y1,x2,y2) return _sq((x2 - x1) ^ 2 + (y2 - y1) ^ 2) end

function findAngle(x1, y1, x2, y2)
	local t = _r2d(_at2(y1 - y2, x2 - x1))
	if t < 0 then t = t + 360 end
	return t
end

